@@ -9,16 +9,12 @@ import android.content.Context;
 import android.os.Build;
 import android.util.Log;
 import android.view.inputmethod.TextAttribute;
-import android.view.WindowMetrics;
 import android.view.inputmethod.BaseInputConnection;
 import android.view.inputmethod.CompletionInfo;
 import android.view.inputmethod.ExtractedText;
 import android.view.inputmethod.ExtractedTextRequest;
 import android.view.inputmethod.InputMethodManager;
 import android.view.KeyEvent;
-import android.graphics.Rect;
-import android.app.Activity;
-import android.util.DisplayMetrics;
 
 class QtExtractedText
 {
@@ -70,6 +66,7 @@ class QtInputConnection extends BaseInputConnection
 
     private static final String QtTAG = "QtInputConnection";
 
+    private boolean m_duringBatchEdit = false;
     private final QtInputConnectionListener m_qtInputConnectionListener;
 
     class HideKeyboardRunnable implements Runnable {
@@ -109,10 +106,12 @@ class QtInputConnection extends BaseInputConnection
 
     private void setClosing(boolean closing)
     {
-        if (closing)
-            m_view.postDelayed(new HideKeyboardRunnable(), KEYBOARD_CHECK_DELAY_MS);
-        else if (m_qtInputConnectionListener != null)
-            m_qtInputConnectionListener.onSetClosing(false);
+        if (android.os.Build.VERSION.SDK_INT < Build.VERSION_CODES.R) {
+            if (closing)
+                m_view.postDelayed(new HideKeyboardRunnable(), KEYBOARD_CHECK_DELAY_MS);
+            else if (m_qtInputConnectionListener != null)
+                m_qtInputConnectionListener.onSetClosing(false);
+        }
     }
 
     QtInputConnection(QtEditText targetView, QtInputConnectionListener listener)
@@ -126,7 +125,7 @@ class QtInputConnection extends BaseInputConnection
 
     void restartImmInput()
     {
-        if (QtNativeInputConnection.fullscreenMode()) {
+        if (QtNativeInputConnection.fullscreenMode() && !m_duringBatchEdit) {
             if (m_imm != null)
                 m_imm.restartInput(m_view);
         }
@@ -137,6 +136,7 @@ class QtInputConnection extends BaseInputConnection
     public boolean beginBatchEdit()
     {
         setClosing(false);
+        m_duringBatchEdit = true;
         return QtNativeInputConnection.beginBatchEdit();
     }
 
@@ -153,7 +153,12 @@ class QtInputConnection extends BaseInputConnection
     public boolean endBatchEdit()
     {
         setClosing(false);
-        return QtNativeInputConnection.endBatchEdit();
+        boolean ret = QtNativeInputConnection.endBatchEdit();
+        if (m_duringBatchEdit) {
+            m_duringBatchEdit = false;
+            restartImmInput();
+        }
+        return ret;
     }
 
     @Override
@@ -167,15 +172,18 @@ class QtInputConnection extends BaseInputConnection
     public boolean commitText(CharSequence text, int newCursorPosition)
     {
         setClosing(false);
+        boolean result = QtNativeInputConnection.commitText(text.toString(), newCursorPosition);
         restartImmInput();
-        return QtNativeInputConnection.commitText(text.toString(), newCursorPosition);
+        return result;
     }
 
     @Override
     public boolean deleteSurroundingText(int leftLength, int rightLength)
     {
         setClosing(false);
-        return QtNativeInputConnection.deleteSurroundingText(leftLength, rightLength);
+        boolean result = QtNativeInputConnection.deleteSurroundingText(leftLength, rightLength);
+        restartImmInput();
+        return result;
     }
 
     @Override
@@ -308,7 +316,9 @@ class QtInputConnection extends BaseInputConnection
     public boolean setComposingText(CharSequence text, int newCursorPosition)
     {
         setClosing(false);
-        return QtNativeInputConnection.setComposingText(text.toString(), newCursorPosition);
+        boolean result = QtNativeInputConnection.setComposingText(text.toString(), newCursorPosition);
+        restartImmInput();
+        return result;
     }
 
     @TargetApi(33)
@@ -351,6 +361,8 @@ class QtInputConnection extends BaseInputConnection
     public boolean setSelection(int start, int end)
     {
         setClosing(false);
-        return QtNativeInputConnection.setSelection(start, end);
+        boolean result = QtNativeInputConnection.setSelection(start, end);
+        restartImmInput();
+        return result;
     }
 }
