@@ -20,6 +20,7 @@
 #include <QtGui/private/qguiapplication_p.h>
 #include <QtGui/private/qwindow_p.h>
 #include <QtGui/private/qapplekeymapper_p.h>
+#include <QtGui/private/qpointingdevice_p.h>
 #include <qpa/qwindowsysteminterface_p.h>
 
 Q_LOGGING_CATEGORY(lcQpaTablet, "qt.qpa.input.tablet")
@@ -58,25 +59,8 @@ inline ulong getTimeStamp(UIEvent *event)
     UIPanGestureRecognizer *m_scrollGestureRecognizer;
     CGPoint m_lastScrollCursorPos;
     CGPoint m_lastScrollDelta;
-}
-
-+ (void)load
-{
-#if !defined(Q_OS_TVOS) && !defined(Q_OS_VISIONOS)
-    if (QOperatingSystemVersion::current() < QOperatingSystemVersion(QOperatingSystemVersion::IOS, 11)) {
-        // iOS 11 handles this though [UIView safeAreaInsetsDidChange], but there's no signal for
-        // the corresponding top and bottom layout guides that we use on earlier versions. Note
-        // that we use the _will_ change version of the notification, because we want to react
-        // to the change as early was possible. But since the top and bottom layout guides have
-        // not been updated at this point we use asynchronous delivery of the event, so that the
-        // event is processed by QtGui just after iOS has updated the layout margins.
-        [[NSNotificationCenter defaultCenter] addObserverForName:UIApplicationWillChangeStatusBarFrameNotification
-            object:nil queue:[NSOperationQueue mainQueue] usingBlock:^(NSNotification *) {
-                for (QWindow *window : QGuiApplication::allWindows())
-                    QWindowSystemInterface::handleSafeAreaMarginsChanged<QWindowSystemInterface::AsynchronousDelivery>(window);
-            }
-        ];
-    }
+#if QT_CONFIG(tabletevent)
+    UIHoverGestureRecognizer *m_hoverGestureRecognizer;
 #endif
 }
 
@@ -118,6 +102,13 @@ inline ulong getTimeStamp(UIEvent *event)
         m_lastScrollCursorPos = CGPointZero;
         [self addGestureRecognizer:m_scrollGestureRecognizer];
 
+#if QT_CONFIG(tabletevent)
+        m_hoverGestureRecognizer = [[UIHoverGestureRecognizer alloc]
+                                     initWithTarget:self
+                                     action:@selector(handleHover:)];
+        [self addGestureRecognizer:m_hoverGestureRecognizer];
+#endif
+
         // Set up layer
         if ([self.layer isKindOfClass:CAMetalLayer.class]) {
             QWindow *window = self.platformWindow->window();
@@ -366,12 +357,12 @@ inline ulong getTimeStamp(UIEvent *event)
     [super traitCollectionDidChange: previousTraitCollection];
 
     QPointingDevice *touchDevice = QIOSIntegration::instance()->touchDevice();
-    QPointingDevice::Capabilities touchCapabilities = touchDevice->capabilities();
+    auto *devicePriv = QPointingDevicePrivate::get(touchDevice);
 
-    touchCapabilities.setFlag(QPointingDevice::Capability::Pressure,
-                              (self.traitCollection.forceTouchCapability == UIForceTouchCapabilityAvailable));
-
-    touchDevice->setCapabilities(touchCapabilities);
+    auto capabilities = touchDevice->capabilities();
+    capabilities.setFlag(QPointingDevice::Capability::Pressure,
+        (self.traitCollection.forceTouchCapability == UIForceTouchCapabilityAvailable));
+    devicePriv->setCapabilities(capabilities);
 }
 
 -(BOOL)pointInside:(CGPoint)point withEvent:(UIEvent *)event
@@ -381,6 +372,39 @@ inline ulong getTimeStamp(UIEvent *event)
     return [super pointInside:point withEvent:event];
 }
 
+#if QT_CONFIG(tabletevent)
+- (void)handlePencilEventForLocationInView:(CGPoint)locationInView withState:(QEventPoint::State)state withTimestamp:(ulong)timeStamp
+    withForce:(CGFloat)force withMaximumPossibleForce:(CGFloat)maximumPossibleForce withZOffset:(CGFloat)zOffset
+    withAzimuthUnitVector:(CGVector)azimuth withAltitudeAngleRadian:(CGFloat)altitudeAngleRadian
+{
+    QIOSIntegration *iosIntegration = QIOSIntegration::instance();
+
+    QPointF localViewPosition = QPointF::fromCGPoint(locationInView);
+    QPoint localViewPositionI = localViewPosition.toPoint();
+    QPointF globalScreenPosition = self.platformWindow->mapToGlobal(localViewPositionI) +
+            (localViewPosition - localViewPositionI);
+    qreal pressure = 0;
+    if (force != 0 && maximumPossibleForce != 0)
+        pressure = force / maximumPossibleForce;
+    // azimuth unit vector: +x to the right, +y going downwards
+    // altitudeAngleRadian given in radians, pi / 2 is with the stylus perpendicular to the iPad, smaller values mean more tilted, but never negative.
+    // Convert to degrees with zero being perpendicular.
+    qreal altitudeAngle = 90 - qRadiansToDegrees(altitudeAngleRadian);
+    qreal xTilt = qBound(-60.0, altitudeAngle * azimuth.dx, 60.0);
+    qreal yTilt = qBound(-60.0, altitudeAngle * azimuth.dy, 60.0);
+
+    qCDebug(lcQpaTablet) << ":" << timeStamp << localViewPosition << pressure << state << "azimuth" << azimuth.dx << azimuth.dy
+                << "altitude" << altitudeAngleRadian << "xTilt" << xTilt << "yTilt" << yTilt;
+    QWindowSystemInterface::handleTabletEvent(self.platformWindow->window(), timeStamp,
+            // device, local, global
+            iosIntegration->pencilDevice(), localViewPosition, globalScreenPosition,
+            // buttons
+            state == QEventPoint::State::Released ? Qt::NoButton : Qt::LeftButton,
+            // pressure, xTilt, yTilt, tangentialPressure, rotation, z, modifiers
+            pressure, xTilt, yTilt, 0, 0, zOffset, Qt::NoModifier);
+}
+#endif
+
 - (void)handleTouches:(NSSet *)touches withEvent:(UIEvent *)event withState:(QEventPoint::State)state withTimestamp:(ulong)timeStamp
 {
     QIOSIntegration *iosIntegration = QIOSIntegration::instance();
@@ -389,31 +413,11 @@ inline ulong getTimeStamp(UIEvent *event)
 #if QT_CONFIG(tabletevent)
     if (m_activePencilTouch && [touches containsObject:m_activePencilTouch]) {
         NSArray<UITouch *> *cTouches = [event coalescedTouchesForTouch:m_activePencilTouch];
-        int i = 0;
         for (UITouch *cTouch in cTouches) {
-            QPointF localViewPosition = QPointF::fromCGPoint([cTouch preciseLocationInView:self]);
-            QPoint localViewPositionI = localViewPosition.toPoint();
-            QPointF globalScreenPosition = self.platformWindow->mapToGlobal(localViewPositionI) +
-                    (localViewPosition - localViewPositionI);
-            qreal pressure = cTouch.force / cTouch.maximumPossibleForce;
-            // azimuth unit vector: +x to the right, +y going downwards
-            CGVector azimuth = [cTouch azimuthUnitVectorInView: self];
-            // azimuthAngle given in radians, zero when the stylus points towards +x axis; converted to degrees with 0 pointing straight up
-            qreal azimuthAngle = qRadiansToDegrees([cTouch azimuthAngleInView: self]) + 90;
-            // altitudeAngle given in radians, pi / 2 is with the stylus perpendicular to the iPad, smaller values mean more tilted, but never negative.
-            // Convert to degrees with zero being perpendicular.
-            qreal altitudeAngle = 90 - qRadiansToDegrees(cTouch.altitudeAngle);
-            qCDebug(lcQpaTablet) << i << ":" << timeStamp << localViewPosition << pressure << state << "azimuth" << azimuth.dx << azimuth.dy
-                     << "angle" << azimuthAngle << "altitude" << cTouch.altitudeAngle
-                     << "xTilt" << qBound(-60.0, altitudeAngle * azimuth.dx, 60.0) << "yTilt" << qBound(-60.0, altitudeAngle * azimuth.dy, 60.0);
-            QWindowSystemInterface::handleTabletEvent(self.platformWindow->window(), timeStamp, localViewPosition, globalScreenPosition,
-                    // device, pointerType, buttons
-                    int(QInputDevice::DeviceType::Stylus), int(QPointingDevice::PointerType::Pen), state == QEventPoint::State::Released ? Qt::NoButton : Qt::LeftButton,
-                    // pressure, xTilt, yTilt
-                    pressure, qBound(-60.0, altitudeAngle * azimuth.dx, 60.0), qBound(-60.0, altitudeAngle * azimuth.dy, 60.0),
-                    // tangentialPressure, rotation, z, uid, modifiers
-                    0, azimuthAngle, 0, 0, Qt::NoModifier);
-            ++i;
+            [self handlePencilEventForLocationInView:[cTouch preciseLocationInView:self] withState:state withTimestamp:timeStamp
+                withForce:cTouch.force withMaximumPossibleForce:cTouch.maximumPossibleForce withZOffset:0
+                withAzimuthUnitVector:[cTouch azimuthUnitVectorInView:self]
+                withAltitudeAngleRadian:cTouch.altitudeAngle];
         }
     }
 #endif
@@ -779,6 +783,31 @@ inline ulong getTimeStamp(UIEvent *event)
 }
 #endif // QT_CONFIG(wheelevent)
 
+#if QT_CONFIG(tabletevent)
+- (void)handleHover:(UIHoverGestureRecognizer *)recognizer
+{
+    if (!self.platformWindow)
+        return;
+
+    ulong timeStamp = [[NSProcessInfo processInfo] systemUptime] * 1000;
+
+    CGFloat zOffset = 0;
+    if (@available(ios 16.1, *))
+        zOffset = [recognizer zOffset];
+
+    CGVector azimuth;
+    CGFloat altitudeAngleRadian = 0;
+    if (@available(ios 16.4, *)) {
+        azimuth = [recognizer azimuthUnitVectorInView:self];
+        altitudeAngleRadian = recognizer.altitudeAngle;
+    }
+
+    [self handlePencilEventForLocationInView:[recognizer locationInView:self] withState:QEventPoint::State::Released
+        withTimestamp:timeStamp withForce:0 withMaximumPossibleForce:0 withZOffset:zOffset
+        withAzimuthUnitVector:azimuth withAltitudeAngleRadian:altitudeAngleRadian];
+}
+#endif
+
 @end
 
 @implementation UIView (QtHelpers)
